/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*                                                                 */
/*               SUBFIELDS OF A NUMBER FIELD                       */
/*                                                                 */
/*   J. Kl\"uners and M. Pohst, J. Symb. Comp. (1996), vol. 11     */
/*                                                                 */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

#include "genpari.h"

static long TR; /* nombre de changements de polynomes (degre fixe) */

GEN potential_block_systems(long N,GEN dd,GEN p,GEN ff,long *n);
GEN calc_block(long N,GEN Z,long d,GEN Y,GEN vbs);
GEN print_block_system(long N,GEN Y,long d);
GEN candidates_for_subfields(GEN nf,GEN A,GEN DATA,GEN *ptdelta,GEN *ptrootsA);
GEN embedding_of_potential_subfields(GEN nf,GEN g,GEN DATA,GEN rootsA,GEN delta);
GEN subfields_of_given_degree(GEN nf,GEN dpol,long d);
GEN subfields(GEN nf,GEN d);
GEN subfieldsall(GEN nf);
GEN Fpn(GEN p,long n);
GEN interpolation_polynomial_mod_p(GEN T,GEN delta,GEN p);
GEN newton_lift(GEN f,GEN g,GEN gp,GEN wh0,GEN pp);
GEN retrieve_p_adique_polynomial_in_Q(GEN nf,GEN h);
GEN bound_for_coeff(GEN pol,long m,GEN rr);
GEN hensel_lift_one_step_for_two(GEN C,GEN A,GEN B,GEN U,GEN V,GEN fkk,GEN p);
GEN hensel_lift_for_two(GEN C,GEN A,GEN B,GEN U,GEN V,GEN fkk,GEN p,long e);
GEN hensel_lift(GEN pol,GEN fk,GEN fkk,GEN p,long e);
GEN special_lift(GEN a,GEN fq);
long exponent_of_potential_block(GEN A,GEN perm);
long test_coefficients(GEN g,GEN M);
GEN compute_data(GEN nf,GEN ff,GEN p,long m);
GEN choose_prime(GEN nf,GEN dpol,long lowerp,long d,GEN *ptff,GEN *ptlistpotbl);
GEN im_by_perm(GEN in,GEN cy,long l);
GEN cycle_power_to_perm(long N,GEN cy,long l);
GEN perm_mul(long N,GEN perm1,GEN perm2);
GEN one_cycle_to_perm(long N,GEN cy);
GEN im_block_by_perm(GEN D,GEN perm);
GEN cycles_to_perm(long N,GEN cys);
long block_egal(GEN A,GEN B);
long in_what_cycle(GEN a,GEN cys);
long what_number(GEN a,GEN cy);

/* Field is given by initalg() */

/* Computation of potential block systems of given size d
   associated to a rational prime p : give a row vector of row vectors
   containing the potential block systems of imprimitivity ; a potential
   block system is a vector of row vectors (enumeration of the roots).
   Give also as a pointer the factorization mod p of the polynomial nf[1] */

GEN
potential_block_systems(long N,GEN dd,GEN p,GEN ff,long *n)
{
  long av,tetpil,r,i,j,k,d;
  GEN p1,p2,Y,Z,vbs;
  
  av=avma;vbs=cgetg(1,17);d=itos(dd);r=lg(ff)-1;Z=cgetg(r+1,17);k=0;
  for(i=1;i<=r;i++)
  {Z[i]=lgetg(n[i]+1,17);p1=(GEN)Z[i];for(j=1;j<=n[i];j++) p1[j]=lstoi(k+j);k+=n[i];}
  Y=cgetg(1,17);p2=calc_block(N,Z,d,Y,vbs);
  tetpil=avma;return gerepile(av,tetpil,gcopy(p2));
}

GEN
calc_block(long N,GEN Z,long d,GEN Y,GEN vbs)
{
  long av,tetpil,r,*n,*non,lK,i,j,k,*pn,*pnon,t,tp,T,lpn,*e,u,nn,nZ,lY,lZp,nd,le;
  GEN K,Zp,Zpp,Yp,a;

  if(DEBUGLEVEL>3)
  {
    fprintferr("avma = %ld\n",avma);
    fprintferr("Z = ");bruterr(Z,'g',-1);fprintferr("\n");
    fprintferr("Y = ");bruterr(Y,'g',-1);fprintferr("\n");
    fprintferr("vbs = ");bruterr(vbs,'g',-1);fprintferr("\n");flusherr();
  }
  r=lg(Z)-1;
  n=(long*)malloc(sizeof(long)*(r+1));
  if(!((long)n)) err(talker,"not allocated memory in calc_block");
  for(i=1;i<=r;i++) n[i]=lg((GEN)Z[i])-1;
  non=(long*)malloc(sizeof(long)*(r+1));
  if(!((long)non)){free(n);err(talker,"not allocated memory in calc_block");}
  pnon=(long*)malloc(sizeof(long)*(r+1));
  if(!((long)pnon)){free(n);free(non);err(talker,"not allocated memory in calc_block");}
  pn=(long*)malloc(sizeof(long)*(r+1));
  if(!((long)pn))
  {free(n);free(non);free(pnon);err(talker,"not allocated memory in calc_block");}
  for(i=1;i<=r;i++) pn[i]=0;
  av=avma;K=divisors(stoi(n[1]));lK=lg(K)-1;
  for(i=1;i<=lK;i++)
  {
    k=itos((GEN)K[i]);for(j=1;j<=r;j++) pnon[j]=0;
    lpn=0;for(j=2;j<=r;j++){if(!(n[j]%k)){lpn++;pn[lpn]=n[j];pnon[lpn]=j;}}
    if(lpn>31)
    {avma=av;free(pn);free(pnon);free(non);free(n);err(talker,"overflow in calc_block");}
    if(!lpn)
    {
      if(!(d*k-n[1]))
      {
	e=(long*)malloc(sizeof(long)*2);
	if(!((long)e))
	{
	  avma=av;free(pn);free(pnon);free(non);free(n);
	  err(talker,"not allocated memory in calc_block");
	}
	T=1;le=1;
      }
      else T=0;
    }
    else
    {
      e=(long*)malloc(sizeof(long)*(lpn+1));
      if(!((long)e))
      {
	avma=av;free(pn);free(pnon);free(non);free(n);
	err(talker,"not allocated memory in calc_block");
      }
      T=itos(gpuigs(gdeux,lpn));le=lpn;
    }
    for(t=0;t<T;t++)
    {
      for(j=1;j<=r;j++) non[j]=0;
      nd=0;tp=t;for(u=1;u<=le;u++){if(tp&1){nd++;e[u]=1;tp>>=1;}else{e[u]=0;tp>>=1;}}
      nn=n[1];for(u=1;u<=le;u++){if(e[u]) nn+=pn[u];}      
      if(!(d*k-nn))
      {
	lZp=nd+1;Zp=cgetg(lZp+1,17);Zp[1]=lcopy((GEN)Z[1]);non[1]=1;nZ=1;
	for(j=1;j<=le;j++)if(e[j]){nZ++;Zp[nZ]=lcopy((GEN)Z[pnon[j]]);non[pnon[j]]=1;}
	lY=lg(Y)-1;Yp=cgetg(lY+2,17);for(j=1;j<=lY;j++) Yp[j]=lcopy((GEN)Y[j]);
	Yp[lY+1]=lcopy(Zp);
	if(gegal(Z,Zp))
	{
	  a=print_block_system(N,Yp,d);
	  vbs=concat(vbs,a);
	}
	else
	{
	  Zpp=cgetg(r-lZp+1,17);u=0;
	  for(j=1;j<=r;j++){if(!non[j]){u++;Zpp[u]=lcopy((GEN)Z[j]);}}
	  vbs=calc_block(N,Zpp,d,Yp,vbs);
	}
	Yp=cgetg(lY+1,17);for(j=1;j<=lY;j++) Yp[j]=lcopy((GEN)Y[j]);
      }
    }
    if(T) free(e);
  }
  free(n);free(non);free(pnon);free(pn);
  tetpil=avma;
  return gerepile(av,tetpil,gcopy(vbs));
}

GEN
print_block_system(long N,GEN Y,long d)
{
  long av,av1,tetpil,i,j,l,ll,fl,r,*s,*k,*kk,*n,lp,**e,u,v,*t,ns,ls;
  GEN D,De,Yi,Z,a,aa,vaa,perm,vbs,Dn,vDn,p1;

  if(DEBUGLEVEL>3){fprintferr("Y = ");bruterr(Y,'g',-1);fprintferr("\n");}
  av=avma;r=lg(Y)-1;D=cgetg(1,17);
  k=(long*)malloc(sizeof(long)*(r+1));
  if(!((long)k)) err(talker,"not allocated memory in print_block_system");
  s=(long*)malloc(sizeof(long)*(r+1));
  if(!((long)s)){free(k);err(talker,"not allocated memory in print_block_system");}
  for(i=1;i<=r;i++) s[i]=lg((GEN)Y[i])-1;
  for(i=1;i<=r;i++)
  {
    Yi=(GEN)Y[i];
    n=(long*)malloc(sizeof(long)*(s[i]+1));
    if(!((long)n))
    {free(k);free(s);err(talker,"not allocated memory in print_block_system");}
    for(j=1;j<=s[i];j++) n[j]=lg((GEN)Yi[j])-1;
    k[i]=n[1];for(j=2;j<=s[i];j++) k[i]+=n[j];k[i]=k[i]/d;
    De=cgetg(k[i]+1,17);for(j=1;j<=k[i];j++) De[j]=lgetg(1,17);
    a=cgetg(s[i]+1,17);for(j=1;j<=s[i];j++) a[j]=((GEN)Yi[j])[1];
    for(j=1;j<=s[i];j++)
    {
      for(l=1;l<=n[j];l++)
      {
	aa=im_by_perm((GEN)a[j],(GEN)Yi[j],l);vaa=cgetg(2,17);vaa[1]=(long)aa;
	lp=l%k[i];if(!lp) lp=k[i];De[lp]=(long)concat((GEN)De[lp],vaa);
      }
    }
    D=concat(D,De);free(n);
  }
  if(DEBUGLEVEL>3)
  {
    fprintferr("D = ");bruterr(D,'g',-1);fprintferr("\n");
    for(i=1;i<=r;i++) fprintferr("s[%ld] = %ld , ",i,s[i]);fprintferr("\n");
    for(i=1;i<=r;i++) fprintferr("k[%ld] = %ld , ",i,k[i]);fprintferr("\n");
    flusherr();
  }
  vbs=cgetg(1,17);
  ns=0;for(i=1;i<=r;i++){if((s[i]>1)&&(k[i]>1)) ns++;}
  if(DEBUGLEVEL>3){fprintferr("ns = %ld\n",ns);flusherr();}
  if(!ns)
  {
    vDn=cgetg(2,17);vDn[1]=(long)D;vbs=concat(vbs,vDn);
    free(k);free(s);tetpil=avma;return gerepile(av,tetpil,gcopy(vbs));
  }
  t=(long*)malloc(sizeof(long)*(ns+1));
  if(!((long)t)) err(talker,"not allocated memory in print_block_system");
  kk=(long*)malloc(sizeof(long)*(ns+1));
  if(!((long)kk)) err(talker,"not allocated memory in print_block_system");
  Z=cgetg(ns+1,17);ls=0;
  for(i=1;i<=r;i++)
  {
    if((s[i]>1)&&(k[i]>1))
    {
      ls++;t[ls]=s[i]-1;kk[ls]=k[i];Z[ls]=lgetg(s[i],17);p1=(GEN)Z[ls];
      for(j=2;j<=s[i];j++) p1[j-1]=((GEN)Y[i])[j];
    }
  }
  if(DEBUGLEVEL>3)
  {
    fprintferr("Z = ");bruterr(Z,'g',-1);fprintferr("\n");
    for(i=1;i<=ns;i++) fprintferr("t[%ld] = %ld , ",i,t[i]);fprintferr("\n");
    for(i=1;i<=ns;i++) fprintferr("kk[%ld] = %ld , ",i,kk[i]);fprintferr("\n");
    flusherr();
  }
  e=(long**)malloc(sizeof(long*)*(ns+1));
  if(!((long)e)) err(talker,"not allocated memory in print_block_system");
  for(i=1;i<=ns;i++)
  {
    e[i]=(long*)malloc(sizeof(long)*(t[i]+1));
    if(!((long)e[i])) err(talker,"not allocated memory in print_block_system");
  }
  for(i=1;i<=ns;i++){for(j=1;j<=t[i];j++) e[i][j]=0;}
  i=ns;
  do
  {
    if(DEBUGLEVEL>3)
    {
      for(l=1;l<=ns;l++)
      {
	for(ll=1;ll<=t[l];ll++)	fprintferr("e[%ld][%ld] = %ld , ",l,ll,e[l][ll]);
	fprintferr("\n");
      }
      fprintferr("\n");flusherr();
    }
    av1=avma;perm=cgetg(N+1,17);for(u=1;u<=N;u++) perm[u]=lstoi(u);
    for(u=1;u<=ns;u++)
    {
      for(v=1;v<=t[u];v++)
	perm=perm_mul(N,perm,cycle_power_to_perm(N,(GEN)((GEN)Z[u])[v],e[u][v]));
    }
    Dn=im_block_by_perm(D,perm);vDn=cgetg(2,17);vDn[1]=(long)Dn;
    if(DEBUGLEVEL>3)
    {fprintferr("Dn = ");bruterr(Dn,'g',-1);fprintferr("\n");flusherr();}
    vbs=concat(vbs,vDn);tetpil=avma;vbs=gerepile(av1,tetpil,gcopy(vbs));
    e[ns][t[ns]]++;
    if(e[ns][t[ns]]>=kk[ns])
    {
      j=t[ns]-1;
      while((j>=1)&&(e[ns][j]==(kk[ns]-1))) j--;
      if(j>=1){e[ns][j]++;for(l=j+1;l<=t[ns];l++) e[ns][l]=0;}
      else
      {
	i=ns-1;fl=1;
	while(fl&&(i>=1))
	{
	  j=t[i];
	  while((j>=1)&&(e[i][j]==(kk[i]-1))) j--;
	  if(j>=1)
	  {
	    fl=0;e[i][j]++;
	    for(l=j+1;l<=t[i];l++) e[i][l]=0;
	    for(ll=i+1;ll<=ns;ll++){for(l=1;l<=t[ll];l++) e[ll][l]=0;}
	  }
	  else i--;
	}
      }
    }
  }
  while(i>=1);
  free(k);free(s);for(i=1;i<=ns;i++) free(e[i]);free(e);free(t);free(kk);
  tetpil=avma;return gerepile(av,tetpil,gcopy(vbs));
}

GEN
candidates_for_subfields(GEN nf,GEN A,GEN DATA,GEN *ptdelta,GEN *ptrootsA)
/* Renvoie un polynome g definissant un sous-corps potentiel, ou :
     gzero : si le polynome trouve n'est pas separable,
     gun   : si les coefficients du polynome trouve sont plus grands que la borne M,
     gdeux : si p divise le discriminant de g,
     trois : si le discriminant du corps defini par g a la puissance d ne divise pas
             le discriminant du corps nf,
     quatre: si le discriminant de g est nul. */

{
  long av,av1,tetpil,N,m,r,i,j,l,**no_cy_for_A,**no_roots_for_A,d,*n,pp,e,nn,dec,fl;
  GEN p,pol,cys,p1,Ai,a,tabroots,MM,fhk,bigfq,delta,g,dg,unmodpe,tabrA,dL,dK;
  
  av=avma;pol=(GEN)DATA[1];N=lgef(pol)-3;m=lg(A)-1;
  p=(GEN)DATA[2];pp=itos(p);
  if(N%m) err(talker,"incompatible potential_block_system in candidates_for_subfields");
  d=N/m;r=itos((GEN)DATA[3]);
  n=(long*)malloc(sizeof(long)*(r+1));
  if(!((long)n)){avma=av;err(talker,"not allocated memory in candidates_for_subfields");}
  for(j=1;j<=r;j++) n[j]=itos((GEN)((GEN)DATA[5])[j]);nn=itos((GEN)DATA[6]);  
  no_cy_for_A=(long**)malloc(sizeof(long*)*(m+1));
  if(!((long)no_cy_for_A))
  {avma=av;err(talker,"not allocated memory in candidates_for_subfields");}
  for(i=1;i<=m;i++)
  {
    no_cy_for_A[i]=(long*)malloc(sizeof(long)*(d+1));
    if(!((long)no_cy_for_A[i]))
    {avma=av;err(talker,"not allocated memory in candidates_for_subfields");}
  }
  for(i=1;i<=m;i++){for(j=1;j<=d;j++) no_cy_for_A[i][j]=0;}
  no_roots_for_A=(long**)malloc(sizeof(long*)*(m+1));
  if(!((long)no_roots_for_A))
  {avma=av;err(talker,"not allocated memory in candidates_for_subfields");}
  for(i=1;i<=m;i++)
  {
    no_roots_for_A[i]=(long*)malloc(sizeof(long)*(d+1));
    if(!((long)no_roots_for_A[i]))
    {avma=av;err(talker,"not allocated memory in candidates_for_subfields");}
  }
  for(i=1;i<=m;i++){for(j=1;j<=d;j++) no_roots_for_A[i][j]=0;}
  cys=(GEN)DATA[7];
  for(i=1;i<=m;i++)
  {
    Ai=(GEN)A[i];
    for(j=1;j<=d;j++)
    {
      a=(GEN)Ai[j];no_cy_for_A[i][j]=in_what_cycle(a,cys);
      no_roots_for_A[i][j]=what_number(a,(GEN)cys[no_cy_for_A[i][j]]);
    }
  }
  bigfq=(GEN)DATA[8];
  MM=shifti((GEN)DATA[10],1);e=itos((GEN)DATA[11]);
  unmodpe=gmodulcp(gun,gpuigs(p,e));
  av1=avma;fhk=(GEN)DATA[12];
  tabroots=cgetg(r+1,17);l=0;
  for(j=1;j<=r;j++)
  {
    tabroots[j]=lgetg(n[j]+1,17);p1=(GEN)tabroots[j];
    for(i=1;i<=n[j];i++){l++;p1[i]=lneg(compo((GEN)fhk[l],1));}
  }
  tetpil=avma;tabroots=gerepile(av1,tetpil,gcopy(tabroots));
  if(DEBUGLEVEL>2)
  {
    for(i=1;i<=r;i++)
    {
      for(j=1;j<=n[i];j++)
      {
	fprintferr("tabroots[%ld][%ld] = ",i,j);
	bruterr((GEN)((GEN)tabroots[i])[j],'g',-1);fprintferr("\n");flusherr();
      }
    }
  }
  delta=cgetg(m+1,17);
  for(i=1;i<=m;i++)
  {
    delta[i]=(long)unmodpe;
    for(j=1;j<=d;j++)
    {
      delta[i]=
      lmul((GEN)delta[i],(GEN)((GEN)tabroots[no_cy_for_A[i][j]])[no_roots_for_A[i][j]]);
    }
    p1=(GEN)((GEN)delta[i])[1];
    delta[i]=(long)gmodulcp(lift(lift((GEN)delta[i])),p1);
  }
  if(DEBUGLEVEL>2)
  {
    for(i=1;i<=m;i++)
    {
      fprintferr("delta[%ld] = ",i);bruterr((GEN)delta[i],'g',-1);
      fprintferr("\n");flusherr();
    }
  }
  fl=1;i=1;
  while(fl&&(i<=m))
  {j=i+1;while(fl&&(j<=m)){if(gegal((GEN)delta[i],(GEN)delta[j])) fl=0;else j++;}i++;}
  if(!fl)
  {
    avma=av;free(n);for(i=1;i<=m;i++) free(no_cy_for_A[i]);free(no_cy_for_A);
    for(i=1;i<=m;i++) free(no_roots_for_A[i]);free(no_roots_for_A);return gzero;
  }
  tabrA=cgetg(m+1,17);
  for(i=1;i<=m;i++)
  {
    tabrA[i]=lgetg(d+1,17);p1=(GEN)tabrA[i];
    for(j=1;j<=d;j++){p1[j]=((GEN)tabroots[no_cy_for_A[i][j]])[no_roots_for_A[i][j]];}
  }
  g=unmodpe;for(i=1;i<=m;i++) g=gmul(g,gsub(polx[0],(GEN)delta[i]));
  g=simplify(centerlift(lift(g)));
  if(DEBUGLEVEL>2){fprintferr("pol. founded = ");bruterr(g,'g',-1);fprintferr("\n");}
  fl=test_coefficients(g,(GEN)DATA[10]);
  if(!fl) g=gcopy(gun);
  else
  {
    if(!gcmp0(discsr(g)))
    {
      dL=(GEN)nf[3];dK=gpuigs(discf(g),N/m);
      if(!gdivise(dL,dK)) g=stoi(3);
      else
      {
	dg=discsr(g);
	if(gdivise(dg,p))
	{
	  free(n);for(i=1;i<=m;i++) free(no_cy_for_A[i]);free(no_cy_for_A);
	  for(i=1;i<=m;i++) free(no_roots_for_A[i]);free(no_roots_for_A);
	  avma=av;return gdeux;
	}
      }
    }
    else g=stoi(4);
  } 
  free(n);
  for(i=1;i<=m;i++) free(no_cy_for_A[i]);free(no_cy_for_A);
  for(i=1;i<=m;i++) free(no_roots_for_A[i]);free(no_roots_for_A);
  *ptdelta=delta;*ptrootsA=tabrA;
  tetpil=avma;g=gcopy(g);*ptdelta=gcopy(*ptdelta);*ptrootsA=gcopy(*ptrootsA);
  dec=lpile(av,tetpil,0)>>TWOPOTBYTES_IN_LONG;
  g+=dec;*ptdelta+=dec;*ptrootsA+=dec;return g;
}

GEN
embedding_of_potential_subfields(GEN nf,GEN g,GEN DATA,GEN rootsA,GEN delta)
/* nf est le corps de nombres, g un polynome de Z[x] candidat
   pour definir un sous-corps, p le nombre premier ayant servi a definir
   le potential block system rootsA donne par les racines avec une approximation
   convenable, e est la precision p-adique des elements de rootsA et delta la liste
   des racines de g dans une extension convenable en precision p^e.
   Renvoie un polynome h de Q[x] tel que f divise g o h et donc tel que le couple
   (g,h) definisse un sous-corps, ou bien gzero si rootsA n'est pas un block system */
{
  long av,tetpil,m,d,i,j,k,fl,N,e;
  GEN wh0,wh1,w0,w1,h0,gp,pp,p1,p2,p3,Delta,f,rA,unmodp,p;

  av=avma;m=lg(rootsA)-1;d=lg((GEN)rootsA[1])-1;f=(GEN)nf[1];N=m*d;
  p=(GEN)DATA[2];e=itos((GEN)DATA[11]);gp=deriv(g,varn(g));unmodp=gmodulcp(gun,p);
  Delta=cgetg(N+1,17);k=0;for(j=1;j<=m;j++){for(i=1;i<=d;i++){k++;Delta[k]=delta[j];}}
  rA=cgetg(N+1,17);k=0;
  for(j=1;j<=m;j++){for(i=1;i<=d;i++){k++;rA[k]=((GEN)rootsA[j])[i];}}
  w0=simplify(interpolation_polynomial_mod_p(rA,Delta,p));
  p1=vecbezout(gmul(unmodp,gsubst(gp,varn(gp),w0)),gmul(unmodp,f));
  h0=simplify(lift((GEN)p1[1]));wh0=cgetg(3,17);wh0[1]=(long)w0;wh0[2]=(long)h0;
  if(DEBUGLEVEL>2)
  {
    fprintferr("w = ");bruterr(w0,'g',-1);fprintferr("\n");
    fprintferr("h = ");bruterr(h0,'g',-1);fprintferr("\n");flusherr();
  }
  fl=1;pp=p;
  while(fl>0)
  {
    wh1=newton_lift(f,g,gp,wh0,pp);
    w1=(GEN)wh1[1];w0=(GEN)wh0[1];
    if(DEBUGLEVEL>2)
    {
      fprintferr("w = ");bruterr(w1,'g',-1);fprintferr("\n");
      fprintferr("h = ");bruterr((GEN)wh1[2],'g',-1);fprintferr("\n");flusherr();
    }
    p3=retrieve_p_adique_polynomial_in_Q(nf,gmul(w0,gmodulcp(gun,pp)));
    p2=retrieve_p_adique_polynomial_in_Q(nf,gmul(w1,gmodulcp(gun,gmul(pp,pp))));
    if(DEBUGLEVEL>2)
    {
      fprintferr("Old Q-polynomial = ");bruterr(p3,'g',-1);fprintferr("\n");flusherr();
      fprintferr("New Q-polynomial = ");bruterr(p2,'g',-1);fprintferr("\n");
    }
    if(gegal(p2,p3)) fl=0;
    else if(!test_coefficients(p2,(GEN)DATA[15])) fl=-1;
    else{pp=gsqr(pp);wh0=wh1;}
  }
  if(fl==-1){avma=av;return gzero;}
  p3=gsubst(g,varn(g),p3);
  if(gdivise(p3,f))
  {
    p1=gadd(polx[0],stoi(TR));tetpil=avma;
    return gerepile(av,tetpil,gsubst(p2,varn(p2),p1));
  }
  else{avma=av;return gzero;}
}

GEN
subfields_of_given_degree(GEN nf,GEN dpol,long d)
/* liste des sous corps de degre d du corps de nombres nf */
{
  long av,av2,tetpil,pp,llist,j,lowerp,N;
  GEN listpotbl,p,ff,A,delta,rootsA,CSF,ESF,p1,p2,LSB;
  GEN newf,newnf,newdpol,DATA;

  N=lgef((GEN)nf[1])-3;lowerp=N;
 LAB1:  
  av=avma;LSB=cgetg(1,17);
  p1=setrand(1);
  if(DEBUGLEVEL>4){fprintferr("Hasard = ");bruterr(p1,'g',-1);fprintferr("\n");}
  p=choose_prime(nf,dpol,lowerp,N/d,&ff,&listpotbl);llist=lg(listpotbl)-1;pp=itos(p);
  DATA=compute_data(nf,ff,p,d);
  for(j=1;j<=llist;j++)
  {
    av2=avma;A=(GEN)listpotbl[j];
    if(DEBUGLEVEL>0)
    {
      fprintferr("\n* Potential block # %ld :",j);bruterr(A,'g',-1);
      fprintferr("\n");flusherr();
    }
    CSF=candidates_for_subfields(nf,A,DATA,&delta,&rootsA);
    if(DEBUGLEVEL>0)
    {
      if(gcmp0(CSF))
      {
	fprintferr("non separable polynomial in candidates_for_subfields : ");
	fprintferr("changing the polynomial f ...\n");flusherr();
      }
      else if(gcmp1(CSF))
      {
	fprintferr("too big coeff. for pol. in candidates_for_subfields\n");
	flusherr();
      }
      else if(gegal(CSF,gdeux))
      {
	fprintferr("the prime number divides the discriminant of g : ");
	fprintferr("changing the prime ...\n");flusherr();
      }
      else if(gegal(CSF,stoi(3)))
      {fprintferr("impossible subfield discriminant\n");flusherr();}
      else if(gegal(CSF,stoi(4)))
      {fprintferr("non irreducible polynomial\n");flusherr();}
      else
      {
	fprintferr("candidate for subfield = ");bruterr(CSF,'g',-1);
	fprintferr("\n");flusherr();
      }
    }
    if(gcmp0(CSF))
    {
      TR++;newf=gsubst((GEN)nf[1],varn((GEN)nf[1]),gsub(polx[0],gun));
      if(DEBUGLEVEL>0)
      {fprintferr("new f = ");bruterr(newf,'g',-1);fprintferr("\n");flusherr();}
      newnf=initalg(newf,5);newdpol=discsr(newf);
      return subfields_of_given_degree(newnf,newdpol,d);
    }
    else if(gcmp1(CSF)) avma=av2;
    else if(gegal(CSF,stoi(3))) avma=av2;
    else if(gegal(CSF,stoi(4))) avma=av2;
    else if(gegal(CSF,gdeux)){lowerp=pp+1;avma=av;goto LAB1;}
    else
    {
      ESF=embedding_of_potential_subfields(nf,CSF,DATA,rootsA,delta);
      if(DEBUGLEVEL>0)
      {fprintferr("embedding = ");bruterr(ESF,'g',-1);fprintferr("\n");flusherr();}
      if(!gcmp0(ESF))
      {
	p1=cgetg(3,17);p1[1]=(long)CSF;p1[2]=(long)ESF;p2=cgetg(2,17);p2[1]=(long)p1;
	LSB=concat(LSB,p2);
	tetpil=avma;LSB=gerepile(av2,tetpil,gcopy(LSB));
      }
      else avma=av2;
    }
  }
  tetpil=avma;return gerepile(av,tetpil,gcopy(LSB));
}

GEN
subfields(GEN nf,GEN d)
{
  long av,tetpil,di,N;
  GEN dpol,p1,LSB;
  
  av=avma;N=lgef((GEN)nf[1])-3;di=itos(d);
  if(di==N)
  {
    LSB=cgetg(2,17);LSB[1]=lgetg(3,17);p1=(GEN)LSB[1];p1[1]=nf[1];p1[2]=(long)polx[0];
    return LSB;
  }
  else if(di==1)
  {
    LSB=cgetg(2,17);LSB[1]=lgetg(3,17);p1=(GEN)LSB[1];p1[1]=(long)polx[0];p1[2]=nf[1];
    return LSB;
  }
  else if((di<=0)||(N%di)){LSB=cgetg(1,17);return LSB;}
  else
  {
    TR=0;dpol=gmul((GEN)nf[3],gmul((GEN)nf[4],(GEN)nf[4]));
    p1=subfields_of_given_degree(nf,dpol,di);tetpil=avma;
    return gerepile(av,tetpil,gcopy(p1));
  }
}

GEN
subfieldsall(GEN nf)
{
  long av,av1,tetpil,N,ld,d,i,j,lNLSB;
  GEN dpol,dg,LSB,NLSB,p1,p2;

  av=avma;N=lgef((GEN)nf[1])-3;
  p1=setrand(1);
  if(DEBUGLEVEL>4){fprintferr("Hasard = ");bruterr(p1,'g',-1);fprintferr("\n");}
  if(isprime(stoi(N)))
  {
    LSB=cgetg(3,17);LSB[1]=lgetg(3,17);LSB[2]=lgetg(3,17);
    p1=(GEN)LSB[1];p1[1]=nf[1];p1[2]=(long)polx[0];
    p1=(GEN)LSB[2];p1[1]=(long)polx[0];p1[2]=nf[1];
    tetpil=avma;return gerepile(av,tetpil,gcopy(LSB));
  }
  dpol=gmul(gmul((GEN)nf[4],(GEN)nf[4]),(GEN)nf[3]);dg=divisors(stoi(N));
  if(DEBUGLEVEL>0)
  {
    fprintferr("\n***** Entree dans subfields\n\n");flusherr();
    fprintferr("pol = ");bruterr((GEN)nf[1],'g',-1);fprintferr("\n");
    fprintferr("dpol = ");bruterr(dpol,'g',-1);fprintferr("\n");
    fprintferr("divisors = ");bruterr(dg,'g',-1);fprintferr("\n");flusherr();
  }
  ld=lg(dg)-1;LSB=cgetg(2,17);
  LSB[1]=lgetg(3,17);p1=(GEN)LSB[1];p1[1]=nf[1];p1[2]=(long)polx[0];
  for(i=2;i<ld;i++)
  {
    p1=setrand(1);
    if(DEBUGLEVEL>4){fprintferr("Hasard = ");bruterr(p1,'g',-1);fprintferr("\n");}
    TR=0;av1=avma;d=itos((GEN)dg[i]);
    if(DEBUGLEVEL>0)
    {fprintferr("\n*** Recherche des sous-corps de degre %ld\n\n",N/d);flusherr();}
    NLSB=subfields_of_given_degree(nf,dpol,N/d);
    if(DEBUGLEVEL>0)
    {
      fprintferr("\nListe des sous-corps de degre %ld :\n",N/d);lNLSB=lg(NLSB)-1;
      for(j=1;j<=lNLSB;j++){bruterr((GEN)NLSB[j],'g',-1);fprintferr("\n");}
      flusherr();
    }
    if(lg(NLSB)>1)
    {LSB=concat(LSB,NLSB);tetpil=av1;LSB=gerepile(av1,tetpil,gcopy(LSB));}
    else avma=av1;
  }
  p1=cgetg(2,17);p1[1]=lgetg(3,17);p2=(GEN)p1[1];
  p2[1]=(long)polx[0];p2[2]=nf[1];LSB=concat(LSB,p1);
  if(DEBUGLEVEL>0) fprintferr("\n***** Sortie de subfields\n\n");flusherr();
  tetpil=avma;return gerepile(av,tetpil,gcopy(LSB));
}

GEN
Fpn(GEN p,long n) /* polynome unitaire irreductible de degre n sur F_p[x] */
{
  long av,av1,tetpil,i,*a,fl,j,l,pp;
  GEN pol,vc,unmodp,fpol,p1;

  if(n<=0) err(talker,"non positive degree in Fpn");
  av=avma;unmodp=gmodulcp(gun,p);pp=itos(p);
  if(n==1)
  {pol=gadd(polx[0],gun);tetpil=avma;return gerepile(av,tetpil,gmul(unmodp,pol));}
  a=(long*)malloc(sizeof(long)*(n+2));
  if(!((long)a)) err(talker,"not allocated memory in Fpn");
  a[1]=1;for(i=2;i<=n+1;i++) a[i]=0;fl=1;
  do
  {
    a[n+1]++;
    if(a[n+1]>=pp)
    {
      j=n;while((j>=2)&&(a[j]==pp-1)) j--;
      if(j>=2){a[j]++;for(l=j+1;l<=n+1;l++) a[l]=0;}else fl=0;
    }
    av1=avma;vc=cgetg(n+2,17);vc[1]=un;for(i=2;i<=n+1;i++) vc[i]=lstoi(a[i]);
    pol=gtopoly(vc,0);p1=setrand(1);
    if(DEBUGLEVEL>4){fprintferr("Hasard = ");bruterr(p1,'g',-1);fprintferr("\n");}
    fpol=factmod(pol,p);
    if((lg((GEN)fpol[1])==2)&&(!cmpii(gcoeff(fpol,1,2),gun))) fl=0;else avma=av1;
  }
  while(fl);
  free(a);pol=gmul(unmodp,pol);tetpil=avma;return gerepile(av,tetpil,gcopy(pol));
}

GEN
interpolation_polynomial_mod_p(GEN T,GEN delta,GEN p)
/* polynome d'interpolation de Lagrange P(x) a coefficients dans Z tel que
   P(T[i]) congru a delta[i] mod p pour tout i */
{
  long av,tetpil,i,j,d;
  GEN T1,delta1,P,p1,p2,unmodp;

  av=avma;unmodp=gmodulcp(gun,p);d=lg(T)-1;P=gzero;
  T1=gmul(unmodp,T);delta1=gmul(unmodp,delta);
  for(i=1;i<=d;i++)
  {
    p1=gun;
    for(j=1;j<=i-1;j++) p1=gmul(p1,gsub(polx[0],(GEN)T1[j]));
    for(j=i+1;j<=d;j++) p1=gmul(p1,gsub(polx[0],(GEN)T1[j]));
    p2=gun;
    for(j=1;j<=i-1;j++) p2=gmul(p2,gsub((GEN)T1[i],(GEN)T1[j]));
    for(j=i+1;j<=d;j++) p2=gmul(p2,gsub((GEN)T1[i],(GEN)T1[j]));
    P=gadd(P,gmul((GEN)delta1[i],gdiv(p1,p2)));
  }
  P=lift(lift(P));
  tetpil=avma;return gerepile(av,tetpil,gcopy(P));
}

GEN
newton_lift(GEN f,GEN g,GEN gp,GEN wh0,GEN pp)
/* Etant donne un polynome g de Z[x], de derivee gp, un nombre entier pp,
   et un couple wh0=[w0,h0] de polynomes dans Z[x] tels que :
   h0(x).gp(w0(x)) congru a 1 et g(w0(x)) congru a 0 (mod f,mod pp),
   on renvoie un couple wh1=[w1,h1] de polynomes dans Z[x] satisfaisant
   les memes conditions en remplacant pp par pp^2, avec de plus
   w1 congru a w0 et h1 congru a h0 (mod pp)
   (cf. J. D. Dixon, J. Austral. Math. Soc., Series A, vol.49, 1990, p.445) */
{
  long av,tetpil;
  GEN w0,h0,w1,h1,wh1,unmodpp,ppp,p1;
  
  av=avma;ppp=gmul(pp,pp);unmodpp=gmodulcp(gun,ppp);
  w0=(GEN)wh0[1];h0=(GEN)wh0[2];
  p1=gsub(w0,gmul(h0,gsubst(g,varn(g),w0)));
  w1=lift(lift(gmodulcp(gmul(unmodpp,p1),f)));
  p1=gmul(h0,gsub(gdeux,gmul(h0,gsubst(gp,varn(gp),w1))));
  h1=lift(lift(gmodulcp(gmul(unmodpp,p1),f)));
  wh1=cgetg(3,17);wh1[1]=(long)w1;wh1[2]=(long)h1;
  tetpil=avma;return gerepile(av,tetpil,gcopy(wh1));
}

GEN
retrieve_p_adique_polynomial_in_Q(GEN nf,GEN h)
/* a partir d'un polynome h(x) dont les coefficients sont definis mod p^k,
   et d'un corps de nombres nf, on construit un polynome a coefficients dans Q
   dont les coefficients ont pour approximation p-adique les coefficients de h */
{
  long av,tetpil,i,deg;
  GEN v,ind;

  av=avma;ind=(GEN)nf[4];deg=lgef(h)-3;v=cgetg(deg+2,17);
  for(i=1;i<=deg+1;i++) v[i]=ldiv(centerlift(gmul(ind,compo(h,deg+2-i))),ind);
  tetpil=avma;return gerepile(av,tetpil,gtopoly(v,0));
}

GEN
bound_for_coeff(GEN pol,long m,GEN rr)
{
  long av,tetpil,lrr,i;
  GEN c1,c2,b,M;

  av=avma;lrr=lg(rr)-1;b=gmax(gun,(GEN)rr[1]);
  for(i=2;i<=lrr;i++) b=gmul(b,gmax(gun,(GEN)rr[i]));
  if(m&1){c1=binome(stoi(m-1),m>>1);c2=binome(stoi(m-1),(m>>1)-1);}
  else{c1=binome(stoi(m-1),m>>1);c2=binome(stoi(m-1),m>>1);}
  M=ground(gadd(gmul(c1,b),c2));tetpil=avma;return gerepile(av,tetpil,gcopy(M));
}

GEN
hensel_lift_one_step_for_two(GEN C,GEN A,GEN B,GEN U,GEN V,GEN fkk,GEN p)
/* Algorithmes 3.5.5. H. Cohen page 137 (1995) : renvoie [A1,B1,U1,V1] */
{
  long av,tetpil;
  GEN unmodp,p1,p2,f,t,A0,B0,A1,B1,g,U0,V0,U1,V1,x;

  av=avma;unmodp=gmodulcp(gun,p);
  p1=gdiv(gsub(C,gmul(A,B)),p);f=gmul(p1,unmodp);p2=gdiventres(gmul(V,f),A);
  t=(GEN)p2[1];
  A0=special_lift((GEN)p2[2],fkk);B0=special_lift(gadd(gmul(U,f),gmul(B,t)),fkk);
  A1=gadd(A,gmul(A0,p));B1=gadd(B,gmul(B0,p));
  p1=gdiv(gsub(gun,gadd(gmul(U,A1),gmul(V,B1))),p);g=gmul(p1,unmodp);
  p2=gdiventres(gmul(V,g),A1);t=(GEN)p2[1];
  U0=special_lift(gadd(gmul(U,g),gmul(B1,t)),fkk);V0=special_lift((GEN)p2[2],fkk);
  U1=gadd(U,gmul(U0,p));V1=gadd(V,gmul(V0,p));
  x=cgetg(5,17);x[1]=(long)A1;x[2]=(long)B1;x[3]=(long)U1;x[4]=(long)V1;
  tetpil=avma;return gerepile(av,tetpil,gcopy(x));
}

GEN
hensel_lift_for_two(GEN C,GEN A,GEN B,GEN U,GEN V,GEN fkk,GEN p,long e)
/* renvoie [Ae,Be,Ue,Ve] */
{
  long av,tetpil,fl,ex;
  GEN p1,pp,Ap,Bp,Up,Vp;

  av=avma;fl=1;pp=p;Ap=A;Bp=B;Up=U,Vp=V;ex=1;
  while(fl)
  {
    p1=hensel_lift_one_step_for_two(C,Ap,Bp,Up,Vp,fkk,pp);ex<<=1;
    if(ex>=e) fl=0;
    else{pp=gsqr(pp);Ap=(GEN)p1[1];Bp=(GEN)p1[2];Up=(GEN)p1[3];Vp=(GEN)p1[4];}
  }
  tetpil=avma;return gerepile(av,tetpil,gcopy(p1));  
}

GEN
hensel_lift(GEN pol,GEN fk,GEN fkk,GEN p,long e)
/* lift de Hensel : fk est la factorisation (vecteur des facteurs) du polynome pol
   en x dans le corps fini defini par le polynome en y fkk (polynomes en x dont
   les coefficients sont des polynomes modulo fkk, dont les coefficients sont
   des entiers modulo p); (les coefficients de fkk sont aussi des entiers modulo p);
   on releve cette factorisation a la precision p^e. Cela est equivalent a travailler
   a la precision P^e dans l'extension non ramifiee de Q_p definie par le polynome
   fkk. Renvoie [R_1,...,R_r] les facteurs liftes */
{
  long av,tetpil,r,i,j;
  GEN p1,B0,A,C1,A1,B1,U1,V1,B,R;

  av=avma;C1=pol;r=lg(fk)-1;A=cgetg(r+1,17);
  for(i=1;i<=r;i++) A[i]=(long)special_lift((GEN)fk[i],fkk);
  R=cgetg(r+1,17);
  for(i=1;i<r;i++)
  {
    A1=(GEN)A[i];B1=(GEN)A[i+1];for(j=i+2;j<=r;j++) B1=gmul(B1,(GEN)A[j]);
    B0=(GEN)fk[i+1];for(j=i+2;j<=r;j++) B0=gmul(B0,(GEN)fk[j]);
    p1=vecbezout((GEN)fk[i],B0);
    U1=special_lift((GEN)p1[1],fkk);V1=special_lift((GEN)p1[2],fkk);
    B=hensel_lift_for_two(C1,A1,B1,U1,V1,fkk,p,e);
    R[i]=lcopy((GEN)B[1]);C1=(GEN)B[2];if(i==(r-1)) R[r]=lcopy((GEN)B[2]);
  }
  tetpil=avma;return gerepile(av,tetpil,gcopy(R));
}

GEN
special_lift(GEN a,GEN fq)
/* a est un polynome en x dont les coefficients sont des polynomes modulo fq dont les
   coefficients sont des entiers modulo p (y compris les coefficients de fq.
   Renvoie a sous la forme d'un polynome en x dont les coefficients sonts des polynomes
   modulo fq dont les coefficients sont dans Z */
{
  long av,tetpil,tb,lb,i;
  GEN fqp,b,c;
  
  av=avma;fqp=lift(fq);b=lift(lift(a));tb=typ(b);
  if(tb==1){tetpil=avma;return gerepile(av,tetpil,gmodulcp(b,fqp));}
  if(tb==10)
  {
    lb=lgef(b);c=cgetg(lb,tb);
    c[1]=b[1];for(i=2;i<lb;i++) c[i]=(long)gmodulcp((GEN)b[i],fqp);
    tetpil=avma;return gerepile(av,tetpil,gcopy(c));
  }
  else{avma=av;err(talker,"wrong type in special_lift");}
  return gnil;
}
  
long
exponent_of_potential_block(GEN A,GEN perm) /* min { k > 0 | perm^k(A) = A } */
{
  long av,k,i,lA;
  GEN B,C;

  av=avma;lA=lg(A)-1;B=A;k=1;
  while(k)
  {
    C=cgetg(lA+1,17);for(i=1;i<=lA;i++) C[i]=lcopy((GEN)perm[itos((GEN)B[i])]);
    if(block_egal(C,A)){avma=av;return k;}else{k++;B=C;}
  }
  return 0;
}

long
test_coefficients(GEN g,GEN M)
/* renvoie 0 si l'un des coefficients de g est de module > M ; 1 sinon */
{
  long av,i,fl,lg;

  av=avma;lg=lgef(g)-2;i=1;fl=1;
  while(fl&&(i<=lg)){if(gcmp(gabs((GEN)g[i+1],3),M)>0) fl=0;else i++;}
  avma=av;return fl;
}

GEN
compute_data(GEN nf,GEN ff,GEN p,long m)
/* etant donne nf et p et la factorisation de nf[1] mod p, et le degre m des sous corps
   cherches, cree un vecteur ligne
   a 12 composantes :
   1 : le polynome nf[1],
   2 : le premier p,
   3 : le nombre de facteurs de ff,
   4 : la factorisation ff,
   5 : la longeur des cycles associes (n_1,...,n_r),
   6 : le ppcm q des (n_i),
   7 : les cycles associes,
   8 : le corps F_(p^q),
   9 : les racines de f dans F_(p^q) par facteur de ff,
   10: la borne M pour les sous-corps,
   11: l'exposant e telle que la precision des lifts soit p^e>2.M,
   12: le lift de Hensel a la precision p^e de la factorisation en facteurs lineaires
       de nf[1] dans F_(p^q),
   13: les valeurs absolues des racines de nf[1],
   14: le max des valeurs absolues des racines de nf[1],
   15: la borne de Hadamard pour les coefficients de h(x) tel que g o h = 0 mod nf[1].
   ces donnees sont valides pour nf, p et m (d) donnes.... */
{
  long av,tetpil,i,j,l,r,*n,e,N,pp,d,r1,r2;
  GEN DATA,p1,p2,p3,cys,bigfq,tabroots,MM,fk,rr,dpol;

  av=avma;DATA=cgetg(16,17);DATA[1]=nf[1];DATA[2]=(long)p;r=lg((GEN)ff)-1;
  DATA[3]=lstoi(r);DATA[4]=(long)ff;n=(long *)malloc(sizeof(long)*(r+1));
  if(!((long)n)){avma=av;err(talker,"not allocated compute_data");}
  for(j=1;j<=r;j++) n[j]=lgef((GEN)ff[j])-3;
  DATA[5]=lgetg(r+1,17);p1=(GEN)DATA[5];for(j=1;j<=r;j++) p1[j]=lstoi(n[j]);
  p1=stoi(n[1]);for(i=2;i<=r;i++) p1=glcm(p1,stoi(n[i]));DATA[6]=(long)p1;
  cys=cgetg(r+1,17);l=0;
  for(i=1;i<=r;i++)
  {cys[i]=lgetg(n[i]+1,17);p1=(GEN)cys[i];for(j=1;j<=n[i];j++) p1[j]=lstoi(l+j);l+=n[i];}
  DATA[7]=(long)cys;
  bigfq=Fpn(p,itos((GEN)DATA[6]));polx[1]=lisexpr("y");
  DATA[8]=(long)gsubst(bigfq,varn(bigfq),polx[1]);
  tabroots=cgetg(r+1,17);pp=itos(p);
  for(j=1;j<=r;j++)
  {
    p1=setrand(1);
    if(DEBUGLEVEL>4){fprintferr("Hasard = ");bruterr(p1,'g',-1);fprintferr("\n");}
    p1=(GEN)factmod9((GEN)ff[j],p,(GEN)DATA[8])[1];tabroots[j]=lgetg(n[j]+1,17);
    p2=(GEN)tabroots[j];p2[1]=lneg(compo((GEN)p1[1],1));
    for(i=2;i<=n[j];i++) p2[i]=(long)gpuigs((GEN)p2[i-1],pp);
  }
  DATA[9]=(long)tabroots;
  p1=(GEN)nf[6];
  N=lgef((GEN)nf[1])-3;r1=itos((GEN)((GEN)nf[2])[1]);r2=itos((GEN)((GEN)nf[2])[2]);
  rr=cgetg(N+1,17);
  for(i=1;i<=r1+r2;i++) rr[i]=p1[i];for(i=r1+r2+1;i<=N;i++) rr[i]=p1[i-r2];
  rr=gabs(rr,4);
  DATA[10]=(long)bound_for_coeff((GEN)nf[1],m,rr);
  MM=shifti((GEN)DATA[10],1);e=1;while(cmpii(gpuigs(p,e),MM)<0) e++;
  DATA[11]=lstoi(e);fk=cgetg(N+1,17);l=0;
  for(j=1;j<=r;j++)
  {for(i=1;i<=n[j];i++){l++;fk[l]=lsub(polx[0],(GEN)((GEN)tabroots[j])[i]);}}
  DATA[12]=(long)hensel_lift((GEN)nf[1],fk,(GEN)DATA[8],p,e);
  DATA[13]=(long)rr;
  p3=(GEN)rr[1];for(i=2;i<=N;i++) p3=gmax(p3,(GEN)rr[i]);DATA[14]=(long)p3;
  d=N/m;p1=gsqrt(gpuigs(stoi(N),N-1),4);
  p2=gmul(gpuigs((GEN)DATA[14],d),gsqrt(gpuigs((GEN)DATA[14],N*(N-1)),4));
  dpol=gmul(gmul((GEN)nf[4],(GEN)nf[4]),(GEN)nf[3]);
  DATA[15]=(long)gdiv(gmul(p1,p2),gsqrt(absi(dpol),4));
  if(DEBUGLEVEL>1)
  {
    fprintferr("DATA =\n");
    fprintferr("f = ");bruterr((GEN)DATA[1],'g',-1);fprintferr("\n");flusherr();
    fprintferr("p = ");bruterr((GEN)DATA[2],'g',-1);fprintferr("\n");flusherr();
    fprintferr("r = ");bruterr((GEN)DATA[3],'g',-1);fprintferr("\n");flusherr();
    fprintferr("ff = ");bruterr((GEN)DATA[4],'g',-1);fprintferr("\n");flusherr();
    fprintferr("lcy = ");bruterr((GEN)DATA[5],'g',-1);fprintferr("\n");flusherr();
    fprintferr("ppcm = ");bruterr((GEN)DATA[6],'g',-1);fprintferr("\n");flusherr();
    fprintferr("cys = ");bruterr((GEN)DATA[7],'g',-1);fprintferr("\n");flusherr();
    fprintferr("bigfq = ");bruterr((GEN)DATA[8],'g',-1);fprintferr("\n");flusherr();
    fprintferr("roots = ");bruterr((GEN)DATA[9],'g',-1);fprintferr("\n");flusherr();
    fprintferr("M = ");bruterr((GEN)DATA[10],'g',-1);fprintferr("\n");flusherr();
    fprintferr("e = ");bruterr((GEN)DATA[11],'g',-1);fprintferr("\n");flusherr();
    fprintferr("hensel = ");bruterr((GEN)DATA[12],'g',-1);fprintferr("\n");flusherr();
    fprintferr("|rr| = ");bruterr((GEN)DATA[13],'g',-1);fprintferr("\n");flusherr();
    fprintferr("max|rr| = ");bruterr((GEN)DATA[14],'g',-1);fprintferr("\n");flusherr();
    fprintferr("hadamard bound = ");bruterr((GEN)DATA[15],'g',-1);fprintferr("\n");
    flusherr();
  }
  free(n);tetpil=avma;return gerepile(av,tetpil,gcopy(DATA));
}

GEN
choose_prime(GEN nf,GEN dpol,long lowerp,long d,GEN *ptff,GEN *ptlistpotbl)
{
  long av,tetpil,j,k,pp,oldllist,llist,r,nn,oldnn,*n,dec,N;
  GEN p,oldp,listpotbl,oldlistpotbl,ff,oldff,p3,p1;
  byteptr di;

  av=avma;di=diffptr;pp=*di;N=lgef((GEN)nf[1])-3;
  while((pp<=lowerp)||(gdivise(dpol,stoi(pp)))) pp=pp+(*(++di));p=stoi(pp);
  k=1;oldllist=-1;
  while(k<11)
  {
    p1=setrand(1);
    if(DEBUGLEVEL>4){fprintferr("Hasard = ");bruterr(p1,'g',-1);fprintferr("\n");}
    ff=(GEN)factmod((GEN)nf[1],p)[1];r=lg(ff)-1;
    if((r>1)&&(r<N))
    {
      n=(long*)malloc(sizeof(long)*(r+1));
      if(!((long)n))
      {avma=av;err(talker,"not allocated memory in subfields_of_given_degree");}
      for(j=1;j<=r;j++) n[j]=lgef((GEN)ff[j])-3;
      p3=stoi(n[1]);for(j=2;j<=r;j++) p3=glcm(p3,stoi(n[j]));nn=itos(p3);
      listpotbl=potential_block_systems(N,stoi(d),p,ff,n);llist=lg(listpotbl)-1;
      if(DEBUGLEVEL>0)
      {
	fprintferr("p = %ld , r = %ld , nn = %ld , #p.b.s = %ld\n",pp,r,nn,llist);
	flusherr();
      }
      if(oldllist==-1){oldllist=llist;oldlistpotbl=listpotbl;oldp=p;oldff=ff;oldnn=nn;}
      else
      {
	if((nn<oldnn)||((nn==oldnn)&&(llist<oldllist)))
	{oldllist=llist;oldlistpotbl=listpotbl;oldp=p;oldff=ff;oldnn=nn;}
      }
      free(n);
    }
    k++;do{pp=pp+(*(++di));p=stoi(pp);}while(gdivise(dpol,p));
  }
  if(DEBUGLEVEL>0)
  {
    fprintferr("Nombre premier choisi : p = ");bruterr(oldp,'g',-1);
    fprintferr("\n");flusherr();
    if(DEBUGLEVEL>2)
    {
      fprintferr("Liste des potential block systems de taille %ld :\n",d);
      for(k=1;k<=oldllist;k++){bruterr((GEN)oldlistpotbl[k],'g',-1);fprintferr("\n");}
      fprintferr("\n");flusherr();
    }
  }
  tetpil=avma;p=gcopy(oldp);*ptlistpotbl=gcopy(oldlistpotbl);*ptff=gcopy(oldff);
  dec=lpile(av,tetpil,0)>>TWOPOTBYTES_IN_LONG;p+=dec;*ptlistpotbl+=dec;*ptff+=dec;
  return p;
}

GEN
im_by_perm(GEN in,GEN cy,long l)
/* cy est un cycle; renvoie l'image de l'entier in par la permutation cy^l */
{
  long lcc,lp,k,kp;

  lcc=lg(cy)-1;lp=l%lcc;k=1;while((k<=lcc)&&(cmpii(in,(GEN)cy[k]))) k++;
  if(k>lcc) return gcopy(in);
  else{kp=(k+lp)%lcc;if(!kp) kp=lcc;return gcopy((GEN)cy[kp]);}
}

GEN
cycle_power_to_perm(long N,GEN cy,long l)
/* cy est un cycle; transforme cy^l en une permutation */
{
  long lcy,lp,i,av,tetpil,j;
  GEN perm,p1,p2;
  
  lcy=lg(cy)-1;lp=l%lcy;
  if(!lp){perm=cgetg(N+1,17);for(i=1;i<=N;i++) perm[i]=lstoi(i);return perm;}
  else
  {
    av=avma;p1=one_cycle_to_perm(N,cy);p2=p1;for(j=2;j<=lp;j++) p2=perm_mul(N,p2,p1);
    tetpil=avma;return gerepile(av,tetpil,gcopy(p2));
  }
}

GEN
perm_mul(long N,GEN perm1,GEN perm2) /* produit de deux permutations */
{
  long av,tetpil,i;
  GEN perm;

  av=avma;
  perm=cgetg(N+1,17);for(i=1;i<=N;i++) perm[i]=lcopy((GEN)perm1[itos((GEN)perm2[i])]);
  tetpil=avma;return gerepile(av,tetpil,gcopy(perm));
}
    
GEN
one_cycle_to_perm(long N,GEN cy) /* transforme le cycle cy en une permutation */
{
  long lcy,i,k;
  GEN perm;

  lcy=lg(cy)-1;perm=cgetg(N+1,17);
  for(i=1;i<=N;i++)
  {
    k=1;while((k<=lcy)&&(cmpsi(i,(GEN)cy[k]))) k++;
    if(k>lcy) perm[i]=lstoi(i);
    else{if(k==lcy) perm[i]=lcopy((GEN)cy[1]);else perm[i]=lcopy((GEN)cy[k+1]);}
  }
  return perm;
}

GEN
im_block_by_perm(GEN D,GEN perm) /* image du block system D par la permutation perm */
{
  long av,tetpil,i,j,lb,lcy;
  GEN Dn,cy,p1;

  av=avma;lb=lg(D)-1;Dn=cgetg(lb+1,17);
  for(i=1;i<=lb;i++)
  {
    cy=(GEN)D[i];lcy=lg(cy)-1;Dn[i]=lgetg(lcy+1,17);p1=(GEN)Dn[i];
    for(j=1;j<=lcy;j++) p1[j]=lcopy((GEN)perm[itos((GEN)cy[j])]);
  }
  tetpil=avma;return gerepile(av,tetpil,gcopy(Dn));
}

GEN
cycles_to_perm(long N,GEN cys)
/* transforme un produit de cycles cys=c_1....c_s en une permutation */
{
  long av,tetpil,nbcy,i;
  GEN p1,p2;

  av=avma;nbcy=lg(cys)-1;p1=one_cycle_to_perm(N,(GEN)cys[1]);
  for(i=2;i<=nbcy;i++){p2=one_cycle_to_perm(N,(GEN)cys[i]);p1=perm_mul(N,p1,p2);}
  tetpil=avma;return gerepile(av,tetpil,gcopy(p1));
}

long
block_egal(GEN A,GEN B) /* teste si deux blocks sont egaux ou non */
{
  long lA,lB,av;
  GEN Ap,Bp;

  lA=lg(A)-1;lB=lg(B)-1;if(lA!=lB){err(talker,"non equal lengthes in block_egal");}
  av=avma;Ap=sort(A);Bp=sort(B);
  if(gegal(Ap,Bp)){avma=av;return 1;}else{avma=av;return 0;}
}

long
in_what_cycle(GEN a,GEN cys)
/* rend le numero du cycle dans le support duquel se trouve a */
{
  long lcys,i,k,fl,nk,av;

  av=avma;lcys=lg(cys)-1;k=1;fl=1;
  while(fl&&(k<=lcys))
  {
    nk=lg((GEN)cys[k])-1;i=1;
    while(fl&&(i<=nk))
    {if(!cmpii(a,(GEN)((GEN)cys[k])[i])){fl=0;avma=av;return k;}else i++;}k++;
  }
  if(fl){avma=av;err(talker,"impossible to find the integer in in_what_cycle");}
  return 0;
}

long
what_number(GEN a,GEN cy)
/* si cy=(a_1,a_2,...,a_s) est un cycle, et a un entier, renvoie i t.q. a_i=a */
{
  long i,lcy,fl;

  lcy=lg(cy)-1;i=1;fl=1;while(fl&&(i<=lcy)){if(!cmpii(a,(GEN)cy[i])) fl=0;else i++;}
  if(i>lcy) err(talker,"integer not found in what_number");return i;
}
